#!/usr/bin/env python
"""
  cbmcat targets a selected file on any supported disk, file, or archive
  format and writes its ascii content (for SEQ files) or binary content
  (all other files) to stdout - likely with the intent that said output
  is piped into some other command or redirected to a file.
"""

# customize your own preferred defaults here, or comment it out
arg_defaults = [('binary', False), ('case', 'lower')]

# standard imports
import argparse
import os
import sys

# local import for CBMDisk
import cbmdisk


def cli():
    """ handle command line arguments """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='using CBMdisk/python v' + cbmdisk.__version__ + ' w/ CBMdisk v' + cbmdisk.library_version
                                     )
    parser.add_argument('path', metavar='FILE',
                        help='a disk/file image or archive filename')
    parser.add_argument('--version', action='version', version='%(prog)s v1.0.0; by Elwix/Style',
                        help='show program version/credits')
    parser.add_argument('-b', '--binary', action='store_true', default=False,
                        help='write unconverted bytes even for SEQ files')
    parser.add_argument('-i', '--index', metavar='INDEX', type=int, required=True,
                        help='write the file entry at INDEX to stdout')
    parser.add_argument('-q', '--quiet', action='count',
                        help='suppress some error output')
    parser.add_argument('-C', '--case', choices=['upper', 'lower'], default='lower',
                        help='if file at INDEX is text, convert as if in upper/graphics or lower/upper mode')
    try:
        parser.set_defaults(**dict(arg_defaults))
    except NameError:
        pass
    ret_args = parser.parse_args()
    return ret_args


def process_filename(cbm_disk_name):
    """ open a given filename as a cbmdisk.Disk, handle errors """
    try:
        process_disk(cbmdisk.Disk(cbm_disk_name), cbm_disk_name)
    except cbmdisk.error as exception:
        if exception.message['code'] in ('ReadDiskName', 'CreateDisk'):
            # we expect we can open files with supported extensions and report if we can't; for all other files, report it if we aren't suppressing
            if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) or args.quiet < 1:
                print >> sys.stderr, 'cbmdisk.error: \"{}\" [{}]'.format(cbm_disk_name, exception.message['message'])
        else:
            raise


def process_disk(cbm_disk, cbm_disk_name):
    """ find the file on the cbmdisk.Disk at index and write to stdout """
    assert isinstance(cbm_disk, cbmdisk.Disk)
    global args

    # make sure the indexed file we want actually exists
    if args.index >= len(cbm_disk.files):
        print >> sys.stderr, 'error: \"{}\" has no file at index {} '.format(cbm_disk_name, args.index)
        exit(1)

    # pull the file and write its bytes to stdout
    cbm_file = cbm_disk.files[args.index]

    if cbm_file.type == 'SEQ' and not args.binary:
        sys.stdout.write(cbmdisk.to_ascii(cbm_file.bytes, upper=(args.case == 'upper')))
    else:
        sys.stdout.write(cbm_file.bytes)


if __name__ == '__main__':
    args = cli()

    # ensure stdout is binary on Windows
    if sys.platform == "win32":
        import msvcrt
        msvcrt.setmode(sys.stdout.fileno(), os.O_BINARY)

    # if the input is a file, process it; reject directories and non-existant input
    if os.path.isfile(args.path):
        process_filename(args.path)
    elif os.path.isdir(args.path):
        print >> sys.stderr, 'error: {} must be a file, not a directory'.format(args.path)
        exit(1)
    else:
        print >> sys.stderr, 'error: {} not found'.format(args.path)
        exit(1)

    exit(0)
