#!/usr/bin/env python
"""
  cbmstitch writes a binary file to a disk image in a specific track/sector
  chain specified on the command line. Note that by default the disk image
  is modified in place - use '-o OUT' to write the modified disk to a
  new file of the same type as the source.

  Should work with any CBMdisk supported format with a track/sector layout
  such as d64, d71, d81, and so on.
"""

# standard imports
import argparse
import os
import sys

# local import for CBMDisk
import cbmdisk


def cli():
    """ handle command line arguments """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='using CBMdisk/python v' + cbmdisk.__version__ + ' w/ CBMdisk v' + cbmdisk.library_version
                                     )
    parser.add_argument('path', metavar='FILE',
                        help='the file to be stitched onto a disk')
    parser.add_argument('--version', action='version', version='%(prog)s v1.0.0; by Elwix/Style',
                        help='show program version/credits')
    parser.add_argument('-c', '--chain', required=True,
                        help='track/sector chain (ex 1,1:1,4:1,7:...)')
    parser.add_argument('-i', '--in-disk',  metavar='IN', required=True,
                        help='write the stitched file to the disk named IN')
    parser.add_argument('-o', '--out-disk', metavar='OUT',
                        help='write the modified disk as a new file named OUT')
    parser.add_argument('-q', '--quiet', action='count',
                        help='suppress some error output')
    parser.add_argument('-D', '--omit-dir', action='store_true',
                        help='don''t write a new directory entry for the stitched file')
    parser.add_argument('-O', '--overwrite', action='store_true',
                        help='overwrite allocated sectors instead of exiting')
    ret_args = parser.parse_args()
    return ret_args


# list of suffixes that we expect to be able to process as a track/sector based format
disk_suffix = ('d64', 'd81', 'g64', 'd71', 'nib', 'd80', 'd82')


def chunked(bytebuf):
    """ split bytes into block sized chunks """
    while bytebuf:
        yield bytebuf[:254]
        bytebuf = bytebuf[254:]


def process_filename(cbm_disk_name):
    """ open a given filename as a cbmdisk.Disk, handle errors """
    try:
        process_disk(cbmdisk.Disk(cbm_disk_name), cbm_disk_name)
    except cbmdisk.error as exception:
        if exception.message['code'] in ('ReadDiskName', 'CreateDisk'):
            # we expect we can open files with supported extensions and report if we can't; for all other files, report it if we aren't suppressing
            if cbm_disk_name.lower().endswith(disk_suffix) or args.quiet < 1:
                print >> sys.stderr, 'cbmdisk.error: \"{}\" [{}]'.format(cbm_disk_name, exception.message['message'])
        else:
            raise


def process_disk(cbm_disk, cbm_disk_name):
    """ stitch the file onto the given cbmdisk.Disk """
    assert isinstance(cbm_disk, cbmdisk.Disk)
    global args

    # open and read the file to be stitched
    try:
        with open(os.path.abspath(args.path), mode='rb') as stitch_file:
            stitch_bytes = stitch_file.read()
        stitch_file.close()
    except Exception as exception:
        print >> sys.stderr, ' {}: \"{}\" [{}]'.format(exception.__class__.__name__, os.path.abspath(args.path), exception.message)
        exit(1)

    # split the file bytes into block-sized chunks
    blocks = list(chunked(stitch_bytes))

    # process the chain argument into a list of tracks/sectors /TODO: better check for malformed chain input
    chain = []
    for item in args.chain.split(':'):
        chain.append(tuple(int(x) for x in item.split(',')))

    # we expect there to be as many chain sectors as the block size of the stitched file - fail otherwise
    if len(blocks) != len(chain):
        print >> sys.stderr, 'error: file size ({} blocks) doesn''t equal given number of chain sectors ({})'.format(len(blocks), len(chain))
        exit(1)

    # build up a list of chain links in reverse order; the last block takes a null track and last-byte index
    links = []
    links.insert(0, [0, len(blocks[-1]) + 1])
    # sanity check every track/sector in the chain as we go: is it valid? is it already allocated?
    for tr, sr in reversed(chain):
        if not cbm_disk.is_sector_valid(tr, sr):
            print >> sys.stderr, 'error: {},{} isn''t a valid track/sector'.format(tr, sr)
            exit(1)
        elif not (args.overwrite or cbm_disk.is_sector_free(tr, sr)):
            print >> sys.stderr, 'error: block at {},{} isn''t free'.format(tr, sr)
            exit(1)
        else:
            links.insert(0, (tr, sr))

    # link and set the stitched file blocks, mark as allocated
    for i, block in enumerate(blocks):
        tr, sr = links[i]
        cbm_disk.set_sector_bytes(tr, sr, chr(links[i + 1][0]) + chr(links[i + 1][1]) + block)
        cbm_disk.set_sector_free(tr, sr, False)

    if not args.omit_dir:
        # allocate a new directory entry at the end of the current directory
        file_entry = cbm_disk.files.create(len(cbm_disk.files))

        # the new entry will be a PRG pointing to the stitched file, named using the (track, sector) of the first block of the stitched file
        file_entry.type = 'PRG'
        file_entry.track, file_entry.sector = links[0]
        file_entry.name = cbmdisk.to_petscii('at ' + str(file_entry.track) + '-' + str(file_entry.sector), upper=False)
        file_entry.size = len(blocks)

    # either write the results to a new file as directed, or overwrite the original
    if args.out_disk:
        cbm_disk.save(args.out_disk)
    else:
        cbm_disk.save(cbm_disk_name)


if __name__ == '__main__':
    args = cli()

    # make sure the stitch path exists and is a real file
    if os.path.isdir(args.path):
        print >> sys.stderr, 'error: {} must be a file, not a directory'.format(args.path)
        exit(1)
    elif not os.path.isfile(args.path):
        print >> sys.stderr, 'error: {} not found'.format(args.path)
        exit(1)

    # if the input disk path is a file, process it; reject directories and non-existent input
    if os.path.isfile(args.in_disk):
        process_filename(args.in_disk)
    elif os.path.isdir(args.in_disk):
        print >> sys.stderr, 'error: {} must be a file, not a directory'.format(args.in_disk)
        exit(1)
    else:
        print >> sys.stderr, 'error: {} not found'.format(args.in_disk)
        exit(1)

    exit(0)
