#!/usr/bin/env python
"""
  cbmvrm searches for and optionally neutralizes c64 virii;
  inspired by D64vrm v0.4/Ian Coog.

  By default, only a scan is performed.
"""

# customize your own preferred defaults here, or comment it out
arg_defaults = [('unicode', True)]


# standard imports necessary for everything else
import argparse
import collections
import glob
import math
import os
import string
import sys

# local import for CBMDisk
import cbmdisk


def cli():
    """ handle command line arguments """
    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     epilog='using CBMdisk/python v' + cbmdisk.__version__ + ' w/ CBMdisk v' + cbmdisk.library_version
                                     )
    parser.add_argument('paths', nargs='+', metavar='FILE',
                        help='a disk/file image or archive filename')
    parser.add_argument('--version', action='version', version='%(prog)s v1.0.0; by Elwix/Style',
                        help='show program version/credits')
    parser.add_argument('-c', '--correct', action='store_true',
                        help='with -n, correct file size and bam; zero freed blocks')
    parser.add_argument('-n', '--neutralize', action='store_true',
                        help='neutralize virus and write modified disk')
    parser.add_argument('-q', '--quiet', action='count',
                        help='suppress some (-q) or most (-qq) error output')
    parser.add_argument('-s', '--suffix', metavar='SUFFIX',
                        help='don''t modify disks in place; write new copies to FILE+SUFFIX')
    parser.add_argument('-v', '--verbose', action='count',
                        help='show results for all disks (-v) and all files (-vv)')
    parser.add_argument('-C', '--case', choices=['upper', 'lower'], default='upper',
                        help='output as if in upper/graphics or lower/upper mode')
    parser.add_argument('-D', '--deep', action='store_true',
                        help='scan cbm subdirs and archives recursively')
    parser.add_argument('-R', '--recursive', action='store_true',
                        help='scan subdirectories recursively')
    parser.add_argument('-U', '--unicode', action='store_true', default=False,
                        help='output contents as unicode (from petscii) instead of ascii')
    try:
        parser.set_defaults(**dict(arg_defaults))
    except NameError:
        pass
    ret_args = parser.parse_args()
    return ret_args


def printable(petscii):
    """ convert a petscii string into a printable (unicode or ascii) string """
    global args
    # if in ascii mode (not unicode) we're going to translate non-printable petscii to a question mark
    if 'transascii' not in printable.__dict__:
        printable.transascii = string.maketrans(''.join(chr(x) for x in xrange(0, 0x20)) + ''.join(chr(x) for x in xrange(0x80, 0xa0)), '?'*0x40)
    if args.unicode:
        return cbmdisk.to_unicode(petscii, upper=(args.case == 'upper'), ctrlchrs=True).encode('utf-8')
    else:
        return cbmdisk.to_ascii(petscii.translate(printable.transascii), upper=(args.case == 'upper'))


def get_link(in_cbm, track, sector):
    """ return the link for the given track, sector """
    assert isinstance(in_cbm, cbmdisk.Disk)
    bytes = in_cbm.get_sector_bytes(track, sector)
    return ord(bytes[0]), ord(bytes[1])


def virus_check(cbm_file):
    """ try to match payload signatures, molte grazie for match bytes to Ian Coog; this is definitely not optimized """
    assert isinstance(cbm_file, cbmdisk.File)

    # we'll punt if the file is too small to contain a known virus
    if len(cbm_file.bytes) < 508:
        return cbm_file, None, None

    load_addr = ord(cbm_file.bytes[0]) + ord(cbm_file.bytes[1]) * 256

    # if the load address is above the expected BASIC start, we're process as though it is relocated to BASIC
    if load_addr > 0x0801:
        load_addr = 0x0801

    # try to match signatures; if we match but the file doesn't load to BASIC start, send back None instead of a block size
    offset = 0x0801-load_addr
    if len(cbm_file.bytes) >= 508+offset and ord(cbm_file.bytes[0x06+offset]) == 0x9E:
        if bytes(cbm_file.bytes).find(bytearray([0x32, 0x30, 0x35, 0x39]), 0x07+offset, 0xff+offset) == 0x07+offset \
                and bytes(cbm_file.bytes).find(bytearray([0xDD, 0x10, 0x49, 0xA9]), 0x0e+offset, 0xff+offset) == 0x0E+offset:
            return cbm_file, 'HIV', 2 if load_addr == 0x0801 else None
        elif bytes(cbm_file.bytes).find(bytearray([0xA0, 0x22, 0xB1, 0x2B]), 0x32+offset, 0xff+offset) == 0x32+offset \
                and bytes(cbm_file.bytes).find(bytearray([0x48, 0xC8, 0xC0, 0x2A]), 0x36+offset, 0xff+offset) == 0x36+offset:
            return cbm_file, 'BHP', 8 if load_addr == 0x0801 else None

    # we didn't match a signature, so return a result that indicates no virus
    return cbm_file, None, None


def process_filename(cbm_disk_name):
    """ open a given filename as a cbmdisk.Disk, handle errors """
    try:
        process_disk(cbmdisk.Disk(cbm_disk_name), cbm_disk_name)
    except cbmdisk.error as exception:
        if exception.message['code'] in ('ReadDiskName', 'CreateDisk'):
            # we expect we can open files with supported extensions and report if we can't; for all other files, report it if we aren't suppressing
            if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) or args.quiet < 1:
                print
                print >> sys.stderr, 'cbmdisk.error: \"{}\" [{}]'.format(cbm_disk_name, exception.message['message'])
        else:
            raise


def process_disk(cbm_disk, cbm_disk_name):
    """ search, recursively as needed, the input cbmdisk.Disk or cbmdisk.Files instance for virii. """
    global args, total_infected_files, total_infected_disks, total_neutralized
    deep_into = []

    # get a list of actual files to work with whether the input was a cbmdisk.Disk or a cbmdisk.Files
    if isinstance(cbm_disk, cbmdisk.Disk):
        files = cbm_disk.files
    elif isinstance(cbm_disk, cbmdisk.Files):
        files = cbm_disk
    else:
        return

    # the first pass checks files we think are valid and worth checking (allocated entry, prg, 'readable' bytes) and builds a list for the second pass
    results = []
    found_any = False
    for cbm_file in files:
        assert isinstance(cbm_file, cbmdisk.File)

        # don't process splats, (likely) unallocated directory entries, or non-PRG
        if cbm_file.name == chr(0x0)*16 or cbm_file.splat or cbm_file.type in ('seq', 'usr', 'rel', 'del', '???'):
            continue

        try:
            # make sure we can get the file content and that it is at least large enough to carry payload (~2 blocks)
            if (len(cbm_file.bytes)) < 508:
                continue
        except Exception:
            # if we can't get the file bytes (probably due to a bad/loop chain), mark it and skip
            results.append((cbm_file, None, -1))
            continue

        # for a deep scan, we'll plunge into subdirs (CBM) and archives
        if args.deep and not args.neutralize and ((cbm_file.type == 'cbm' and cbm_disk.type == 'd81') or cbm_file.is_compressed):
            already_added = False
            # some logic to make sure we deep process a 4zip file only one time instead of four
            if cbm_file.name.startswith(('1!', '2!', '3!', '4!')):
                for entry in deep_into:
                    if cbm_file.name[2:] in entry.name:
                        already_added = True
                        break
            if not already_added:
                deep_into.append(cbm_file)

        # check for payload
        result = virus_check(cbm_file)
        results.append(result)
        if result[1]:
            found_any = True

    # results is now a list of 3-tuples for each file: (file, virus type or None, virus size in blocks or None)

    # the second pass outputs file names, status, tallies result counts, and neutralizes virii if requested
    if found_any:
        print
        print '{}: infected'.format(cbm_disk_name)
        total_infected_disks += 1
    elif args.verbose >= 1:
        print
        print '{}: clear'.format(cbm_disk_name)

    if found_any or args.verbose >= 2:
        for result in results:
            printable_file_name = printable(result[0].name)

            if not result[1]:
                # output for a file we couldn't read/check
                if result[2] == -1 and args.verbose >= 2:
                    print '  \"{}\"{: <{width}} *** read fail: not scanned ***'.format(
                        printable_file_name,
                        '',
                        width=16-len(result[0].name)
                    )
                # output for a negative file (we only print negative files w/ -vv)
                elif args.verbose >= 2:
                    print '  \"{}\"{: <{width}} negative'.format(
                        printable_file_name,
                        '',
                        width=16-len(result[0].name)
                    )
            else:
                # we did detect a virus, now we'll loop/neutralize until the file is free of payload
                neutralized = []
                while result[1]:
                    # if the virus block size was reported as None, then we need to write a new file rebased to $0801
                    if not result[2]:
                        load_addr = ord(result[0].bytes[0]) + ord(result[0].bytes[1]) * 256
                        neutralized_bytes = chr(0x01) + chr(0x08) + result[0].bytes[2+0x0801-load_addr:]
                        if cbm_disk.free_blocks < int(math.ceil(len(neutralized_bytes)/254.0)):
                            # if we don't have room for the rebased file, punt
                            neutralized.append(result[1])
                            break
                        else:
                            # create a new file named using the first 15 characters of the infected file, prepended with '+'
                            file_entry = cbm_disk.files.create(len(cbm_disk.files))
                            file_entry.type = result[0].type
                            file_entry.splat = False
                            file_entry.name = ('+'+result[0].name[:15]) if args.neutralize else result[0].name
                            file_entry.size = int(math.ceil(len(neutralized_bytes)/254.0))
                            file_entry.bytes = neutralized_bytes
                            # scan the new file
                            result = virus_check(file_entry)
                    else:
                        # neutralize the virus: relink the original prg then optionally correct the file size and bam
                        tr, sr = result[0].track, result[0].sector
                        for i in range(result[2]):
                            next_tr, next_sr = get_link(cbm_disk, tr, sr)
                            if args.correct:
                                cbm_disk.set_sector_free(tr, sr, True)
                                cbm_disk.set_sector_bytes(tr, sr, chr(0)*256)
                            tr, sr = next_tr, next_sr
                        result[0].track, result[0].sector = tr, sr
                        if args.correct:
                            result[0].size = int(len(result[0].bytes)/254 + 1)
                        neutralized.append(result[1])
                        total_neutralized += 1
                        result = virus_check(result[0])

                # output the final status of the file (how many infections, were they neutralized)
                print '  \"{}\"{: <{width}} {} {}'.format(
                    printable_file_name,
                    '',
                    ('no room on disk, still has' if result[1] else 'neutralized') if args.neutralize else 'has',
                    ', '.join('{} x{}'.format(l[0], l[1]) for l in collections.Counter(neutralized).most_common()),
                    width=16-len(result[0].name)
                )
                total_infected_files += 1

    if args.neutralize and total_neutralized > 0:
        # either write a suffixed copy of the disk, or the original
        if args.suffix:
            cbm_disk.save(os.path.splitext(cbm_disk_name)[0] + args.suffix + os.path.splitext(cbm_disk_name)[1])
        else:
            cbm_disk.save(cbm_disk_name)

    # deep process all necessary file entries
    for cbm_file in deep_into:
        assert isinstance(cbm_file, cbmdisk.File)
        if cbm_file.is_compressed:
            # deep process an archive file
            try:
                process_disk(cbm_file.uncompress()[0], '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)))
            except cbmdisk.error as exception:
                # if we can't uncompress a file that we expected to, output an error if allowed
                if exception.message['code'] in 'FileDecompress':
                    if cbm_disk_name.lower().endswith(tuple(cbmdisk.extensions)) and args.quiet < 2:
                        print
                        print >> sys.stderr, '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)) + ':'
                        print >> sys.stderr, ' {}: [{}]'.format('cbmdisk.error', exception.message['message'])
                else:
                    raise
        elif cbm_file.type == "cbm":
            # deep process a CBM subdirectory
            process_disk(cbm_file.files, '{}{}{}'.format(cbm_disk_name, '|', printable(cbm_file.name)))


if __name__ == '__main__':
    args = cli()

    # if requested, try to find and enable the unicode console module; otherwise ignore unicode argument
    if args.unicode:
        try:
            import win_unicode_console
            win_unicode_console.enable()
        except ImportError:
            args.unicode = False
            sys.exc_clear()

    # glob any path/file arguments as needed
    all_paths = []
    for path in args.paths:
        if any(i in path for i in '*?['):
            all_paths += glob.glob(path)
        else:
            all_paths.append(path)

    # terminate with error if no path/file matched a glob
    if len(all_paths) == 0:
        print >> sys.stderr, 'error: no matching files found'
        exit(1)

    total_infected_files = 0
    total_infected_disks = 0
    total_neutralized = 0

    # process every globbed path/file argument
    traversed = []
    for path in all_paths:
        # if it's a file, process it directly
        if os.path.isfile(path) and path not in traversed:
            process_filename(path)
            traversed.append(path)
        # if it's a directory and we're recursing, walk it
        elif os.path.isdir(path):
            if args.recursive:
                for walk_root, walk_dirnames, walk_filenames in os.walk(path):
                    if walk_root not in traversed:
                        for walk_filename in walk_filenames:
                            process_filename(os.path.join(walk_root, walk_filename))
                            traversed.append(os.path.join(walk_root, walk_filename))
            else:
                print >> sys.stderr, 'error: {} must be a file, not a directory'.format(path)
                exit(1)
        else:
            print >> sys.stderr, 'error: {} not found'.format(path)
            exit(1)

    # pretty print some results
    print
    if total_infected_files:
        print 'found: {:d} infected file{:s>{nf}} on {:d} disk{:s>{nd}}{}'.format(
            total_infected_files,
            '',
            total_infected_disks,
            '',
            '; neutralized {:d} infection{:s>{ni}}'.format(total_neutralized,
                                                           '',
                                                           ni=int(total_neutralized > 1 or total_neutralized == 0)
                                                           ) if args.neutralize else '',
            nf=int(total_infected_files > 1),
            nd=int(total_infected_disks > 1),
            sep=''
        )
    else:
        print 'didn''t find any infected files'

    exit(0)
